//
// Module
const fs = require('fs');  
var BDFParser = require("bdfparser").YBDFParser;  
var NodeHash = require("nodehash-js");
const LineByLine = require('readlinesyn'); 
var  RSM = require ("surrogate").RSM ;
const num = require('numeric-1.2.6.js');  

class match{
    constructor(){
        // nodeData = new NodeHash();
          this._liner = new LineByLine();
          this._rsm =  new RSM();  
          this._bdf = new BDFParser(); 
          this._bdf.init();

          this._bdffile = ""; 
          
          this.bdfjsonfile = "./data/bdfdata.json"
          this._bdfJson = {};

          this._presDataHash = NodeHash; // new NodeHash();
          this._pressureHashJsonFile = "./data/wing-pressure.json";
          this._pressureDataFile = "./wing/Ma6-wing.txt"

          this._newbdfFile = "./new.bdf"

          this._lengthUnitMulti = 1000;
          this._grid_ZeroNeighbor = 0; 


          this._SPC_ID = 2;
          this._LOAD_ID = 2;

          this._SPC1_ID = 1;
          this._pload_ID = 1;
          this._solid_prop_ID = 1;
    }


    saveBdf( theFilePath ){ 
        var self = this;
	
        if(undefined != theFilePath){
            self._newbdfFile = theFilePath;
        }
        if(undefined == self._newbdfFile){
            console.error("saveBdf(): Invalid bdf file path");
            return;
        }

        self._bdf.openTargetFile( self._newbdfFile );

      //
      // Heading 
      var strlist = ["$ MSC.Nastran job created on " + Date(), 
      "$ Linear Static Analysis, Database", 
      "SOL 101", 
      "CEND", 
      "TITLE = Wing", 
      "ECHO = NONE", 
      "SUBCASE 1", 
      "$ Subcase name : Default", 
      "   SUBTITLE=Default", 
      "   SPC=" + self._SPC_ID, 
      "   LOAD=" + self._LOAD_ID , 
      "   DISPLACEMENT(SORT1,REAL)=ALL", 
      "   SPCFORCES(SORT1,REAL)=ALL", 
      "   STRESS(SORT1,REAL,VONMISES,BILIN)=ALL",
      "BEGIN BULK",
      "PARAM    POST    0",
      "PARAM   PRTMAXIM YES" ] 
      self._bdf.writeTargetFile( strlist );

        self._bdf.writeDataVector("MAT1"); 
        self._bdf.writeDataVector("PSOLID"); 
        self._bdf.writeDataVector("PSHELL");         
        self._bdf.writeDataVector('GRID'); 
        self._bdf.writeDataVector('CHEXA');  
        self._bdf.writeDataVector('CPENTA');
        self._bdf.writeDataVector('CQUAD4');
        self._bdf.writeDataVector("CTRIA3");

        self._bdf.writeDataVector("SPCADD");  
        self._bdf.writeDataVector("SPC1"); 
		self._bdf.writeDataVector("LOAD");
        self._bdf.writeDataVector("PLOAD"); 
		
        self._bdf.writeTargetFile("ENDDATA create by @bdfparser" );         
        self._bdf.closeTargetFile( ); 
    }

    openBdfFile( theFilePath ){ 
        var self = this;
   
        if(undefined == theFilePath){
          console.error("openBdfJSONFile() Invalid data file path")
          return;
        } else{
            if(fs.statSync( theFilePath ).isFile()){
                self._bdffile  = theFilePath;
            }else{
                console.error("openBdfJSONFile() Invalid JSON data file path")
                return;
            }
        }
        self._bdf.open( self._bdffile ); 
        return;
    } 

    openPressureFile( theFilePath ){ 
        var self = this;

        if(undefined == theFilePath){
          console.error("openPressureJSONFile() Invalid data file path")
          return;
        }else{
            if(fs.statSync( theFilePath ).isFile()){
                self._pressureDataFile  = theFilePath;
            }else{
                console.error("openPressureJSONFile() Invalid JSON data file path")
                return;
            }
        }
        // self._presDataHash.openJSONFile( self._pressureHashJsonFile ); 
        // Logging
        //console.log("openPressureJSONFile()", self._pressureDataFile );
        self._presDataHash.openDataFile( self._pressureDataFile); 
        return;
    } 

    constraint(){
        var self = this;

        var node_num = 0;

        var p1 = [5000.0, 0.0, -143.693];
        var p2 = [5000.0, 25.0, -143.693];
        var p3 = [3956.367, 0.0, 167.4278];
        var v1 = num.sub( p1, p2);
        var v2 = num.sub( p1, p3);  
        var v3 = self.normalVec( self.crossMulti(v1, v2) ); 
        
        //console.log("v3", v3); 

        var spcDataVector = [ "SPC1", self._SPC1_ID , "  123456" ]
        var card_num = self._bdf.sizeofDataVector( "GRID" ); 
        console.log( "The model has %f GRID elements", card_num ); 
        for( var i = 0; i < 5500; i++ ){
            var theCard = self._bdf.getCardViaSerialNumber( "GRID", i );      
        // var nodes = self._bdf._bdfJson["node"];
        // for( var ite in nodes ){ 
        //     var theNode = nodes[ite];
            var id = theCard[1];
            var x = theCard[3];
            var y = theCard[4];
            var z = theCard[5];

            if( parseFloat(z) > 170.0 )
            {
                continue;
            }
            var p = [x, y, z];
            
            var v = self.normalVec( self.vec( p1, p) );
            //
            var r = self.dot( v3, v );

            // 
            // console.log("constraint(): node", x, y, z); 
            // console.log("constraint(): r", r); 

            if( undefined == r ){ 
                console.log("constraint(): Invalid properities");
                console.log("constraint(): node", x, y, z);                 
                continue;
            }
            //
            if( Math.abs(parseFloat(r)) < 0.00001){
                //console.log("v ", v ); 
                //console.log("r ", r ); 
                var spcID = self._liner.int2FormateStr( parseInt(id), 8);
                spcDataVector.push( spcID ); 
                node_num++;
            }
        }

        console.log("constrained node number: ", node_num);
        console.log("spcDataVector: ", spcDataVector); 
        // self._bdf.addEntry(spcDataVector);
        self._bdf.saveCard( spcDataVector );
		return;
    }


    checkGrid( node_ID ){
        var self = this;

        if( isNaN(node_ID) ){
            return;
        }        
        var p1 = self._bdf.get( "GRID", node_ID ); 
        if(undefined == p1){
            console.log("checkGrid: invalid grid %i", node_ID);
            return;
        }
        var x = parseFloat(p1[3]) / 1000;
        var y = parseFloat(p1[4]) / 1000;
        var z = parseFloat(p1[5]) / 1000;   
        
        if( isNaN(x) ){
            console.log("checkGrid: invalid grid x" );
            return 0.0;
        }  
        if( isNaN(y) ){
            console.log("checkGrid: invalid grid y" );
            return 0.0;
        }  
        if( isNaN(z) ){
            console.log("checkGrid: invalid grid z" );
            return 0.0;
        }  

        var result ;
        try{
            result = self._presDataHash.find( x,y,z);
        }catch(exp){
           console.log(exp);
        }
        if( undefined != result ) { 
            // console.log( r );
            var thejson = JSON.parse( result ); 
            if( undefined == thejson ) { 
                // Logging 
                console.log("checkGrid: (%f, %f, %f)", x, y, z );
                console.log("checkGrid: invalid pressure data:", result  );
                return 0.0;
            }
            var clusterNum = thejson["ClusterNum"]; 
            if( parseInt(clusterNum) == 0 ){ 
                console.log( "Grid %i (%f, %f, %f) has NO neighbor node:", node_ID, x, y, z );
                self._grid_ZeroNeighbor++;
                return;
            }

            // Logging 
            //console.log("result:", JSON.stringify(thejson, null, " "));

            var nodes = thejson["nodes"];
            var cluster = thejson["Cluster"]; 
            // Checking 
            if( undefined == nodes ) { 
                return 0.0;
            }
            if( undefined == cluster ) { 
                return 0.0;
            }

            var theClusterID = cluster[0]; 

            var preData = []; 
            var pressure_sum = 0.0;
 
            for( var ite in cluster ){
              var id = cluster[ite];
              if( parseInt(id) == theClusterID ){
              var x = parseFloat( nodes[ite][0]  );
              var y = parseFloat( nodes[ite][1]  );
              var z = parseFloat( nodes[ite][2]  );  
              var p = parseFloat( nodes[ite][3] );   
              var v = [ [x, y, z], [p]];
              preData.push( v );   
              pressure_sum += p;         
              }
            } 

            if( preData.length < 4 ){
                return (parseFloat( pressure_sum) / preData.length);
            }

            this._rsm.setStandardization(true);
            this._rsm.setRankNumber( 1 );
            this._rsm.setExperimentData( preData );

            var thePre = this._rsm.evaluate( [parseFloat(x), parseFloat(y), parseFloat(z) ] );
            // this._rsm.report(); 
            //console.log('RSM' , thePre );  

            return thePre;
        } 
        return;        
    }

    checkCtria(){
        var self = this;
        self._bdf.updateType( "CTRIA3", 2, 2 );   

        var elm_num = self._bdf.sizeofDataVector( "CTRIA3" ); 
        console.log( "The model has %f CTRIA3 elements", elm_num ); 
        for( var i = 0; i < elm_num; i++ ){
            var theElm = self._bdf.getCardViaSerialNumber( "CTRIA3", i );  
            // Checking the element
            if(undefined == theElm){
                continue;
            }
            var elm_ID = parseInt(theElm[1]); 
            // The nodes 
            var p1_ID = parseInt(theElm[3]); 
            var p2_ID = parseInt(theElm[4]);
            var p3_ID = parseInt(theElm[5]);

            var p1 = self.checkGrid( p1_ID ); 
            var p2 = self.checkGrid( p2_ID );  
            var p3 = self.checkGrid( p3_ID );   

            if( ( undefined == p1 ) || (isNaN(p1) ) ) {
                continue;
            }
            if( ( undefined == p2 ) || (isNaN(p2) ) ) {
                continue;
            }
            if( ( undefined == p3 ) || (isNaN(p3) ) ) {
                continue;
            } 
            var r =  ( p1 + p2 + p3  )/3.0;            
            if( undefined != r ) {
                self._pload_ID;

                // var pressure = self._liner.float2FormateStr( r[0]["pressure"], 8, 2 ); 
                var pressure = self._liner.float2FormateStr( r, 8, 2 );
                //
                // To be inserted the RSM method 

                var theLoad = [ "PLOAD4", self._pload_ID, elm_ID, p1, p2, p3 ]            
                //self._bdf.addEntry("PLOAD", self._pload_ID, theLoad);
                self._bdf.saveCard( theLoad );
                //
                //console.log("checkCtria", theLoad );                
            }            
        }
        return;
    }


    checkcquad4(){
        var self = this;
       
        var elm_num = self._bdf.sizeofDataVector( "CQUAD4" ); 
        console.log( "The model has %f CQUAD4 elements", elm_num );    
        for( var i = 0; i < elm_num; i++ ){
            var theElm = self._bdf.getCardViaSerialNumber( "CQUAD4", i );  
            // Checking the element
            if(undefined == theElm){
                // logging 
                console.log( "checkcquad4(): invalid theElm: ", theElm ); 
                // logging                 
                continue;
            }

            var elm_ID = parseInt(theElm[1]); 
            // The nodes 
            var p1_ID = parseInt(theElm[3]); 
            var p2_ID = parseInt(theElm[4]);
            var p3_ID = parseInt(theElm[5]);
            var p4_ID = parseInt(theElm[6]);

            var p1 = self.checkGrid( p1_ID ); 
            var p2 = self.checkGrid( p2_ID );  
            var p3 = self.checkGrid( p3_ID );  
            var p4 = self.checkGrid( p4_ID ); 

            if( ( undefined == p1 ) || (isNaN(p1) ) ) {
                continue;
            }
            if( ( undefined == p2 ) || (isNaN(p2) ) ) {
                continue;
            }
            if( ( undefined == p3 ) || (isNaN(p3) ) ) {
                continue;
            }    
            if( ( undefined == p4 ) || (isNaN(p4) ) ) {
                continue;
            }             
                // logging 
                //console.log( "checkcquad4() pressure:",  p1, p2, p3, p4 );
                // logging             
            var r =  ( p1 + p2 + p3 + p4 )/4.0;
    
            if( undefined != r ) {


                self._pload_ID;

                // var pressure = self._liner.float2FormateStr( r[0]["pressure"], 8, 2 ); 
                //var pressure = self._liner.float2FormateStr( r , 8, 2 );

                //var theLoad = [ "PLOAD", self._pload_ID, pressure, theElm[3], theElm[4], theElm[5], theElm[6]]            
                var theLoad = [ "PLOAD4", self._pload_ID, elm_ID, p1, p2, p3, p4 ]  
                 // logging 
                 //console.log( "checkcquad4():", theLoad );
                 // logging                
                
                self._bdf.saveCard( theLoad );
            }

        } 
        return;     
    }
 
  //
  //
  normalVec( v ){
    if( !Array.isArray(v) ){
      console.log("Invalid parameters");
      return;
    }

    var lensqr = parseFloat(v[0]) * parseFloat(v[0]) + parseFloat(v[1]) * parseFloat(v[1]) + parseFloat(v[2]) * parseFloat(v[2]) ;
    if( Math.abs(lensqr) < 0.000001 ){
      console.log("Invalid vector length");
      return;
    }
    var len = Math.sqrt( lensqr );
    return [parseFloat(v[0])/parseFloat(len), parseFloat(v[1])/parseFloat(len), parseFloat(v[2])/parseFloat(len) ]
  }
  dot( v1, v2 ){
    if( !Array.isArray(v1) || !Array.isArray(v2)){
        console.log("dot(): Invalid parameters");
        return;
    }
    var result = parseFloat(v1[0]) * parseFloat(v2[0]) 
             + parseFloat(v1[1]) * parseFloat(v2[1]) 
             + parseFloat(v1[2]) * parseFloat(v2[2]) ;
    return result;
  }

  crossMulti( v1, v2 ){
    if( !Array.isArray(v1) || !Array.isArray(v2)){
        console.log("cross(): Invalid parameters");
        return;
    }

    var x = parseFloat(v1[1]) * parseFloat(v2[2]) 
          - parseFloat(v1[2]) * parseFloat(v2[1]); 
    if( isNaN(x) ){
       console.log("invalid result - x");
       return;
    }

    var y = parseFloat(v1[2]) * parseFloat(v2[0]) 
          - parseFloat(v1[0]) * parseFloat(v2[2]);
    if( isNaN(y) ){
       console.log("invalid result - y");
       return;
    }

    var z = parseFloat(v1[0]) * parseFloat(v2[1]) 
          - parseFloat(v1[1]) * parseFloat(v2[0]);
    if( isNaN(z) ){
       console.log("invalid result - z");
       return;
    }
    return [x, y, z];
  } 

  vec( p1, p2 ){
    if( !Array.isArray(p1) || !Array.isArray(p2)){
        console.log("Invalid parameters");
        return;
    }
    var x = parseFloat(p2[0]) - parseFloat(p1[0]); 
    if( isNaN(x) ){
       console.log("invalid result - x");
       return;
    }

    var y = parseFloat(p2[1]) - parseFloat(p1[1]);
    if( isNaN(y) ){
       console.log("invalid result - y");
       return;
    }

    var z = parseFloat(p2[2]) - parseFloat(p1[2]);
    if( isNaN(z) ){
       console.log("invalid result - z");
       return;
    }
    return [x, y, z];
  }
}
module.exports = match;